// Copyright (c) Microsoft Corporation.
// Licensed under the MIT License.

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <myst/file.h>
#include <myst/json.h>

const char* arg0;

typedef struct _callback_data
{
    int depth;
    int newline;
    int comma;
    char* ptr;
    size_t size;
    FILE* stream;
} callback_data_t;

static void _print_string(const char* str, callback_data_t* cd)
{
    fprintf(cd->stream, "\"");

    while (*str)
    {
        char c = *str++;

        switch (c)
        {
            case '"':
                fprintf(cd->stream, "\\\"");
                break;
            case '\\':
                fprintf(cd->stream, "\\\\");
                break;
            case '\b':
                fprintf(cd->stream, "\\b");
                break;
            case '\f':
                fprintf(cd->stream, "\\f");
                break;
            case '\n':
                fprintf(cd->stream, "\\n");
                break;
            case '\r':
                fprintf(cd->stream, "\\r");
                break;
            case '\t':
                fprintf(cd->stream, "\\t");
                break;
            default:
            {
                if (isprint(c))
                    fprintf(cd->stream, "%c", c);
                else
                    fprintf(cd->stream, "\\u%04X", c);
            }
        }
    }

    fprintf(cd->stream, "\"");
}

static void _print_value(
    json_type_t type,
    const json_union_t* value,
    callback_data_t* cd)
{
    switch (type)
    {
        case JSON_TYPE_NULL:
            fprintf(cd->stream, "null");
            break;
        case JSON_TYPE_BOOLEAN:
            fprintf(cd->stream, "%s", value->boolean ? "true" : "false");
            break;
        case JSON_TYPE_INTEGER:
            fprintf(cd->stream, "%ld", value->integer);
            break;
        case JSON_TYPE_REAL:
            fprintf(cd->stream, "%E", value->real);
            break;
        case JSON_TYPE_STRING:
            _print_string(value->string, cd);
            break;
        default:
            break;
    }
}

static void _indent(int depth, callback_data_t* cd)
{
    size_t i;

    for (i = 0; i < depth; i++)
        fprintf(cd->stream, "  ");
}

static json_result_t _callback(
    json_parser_t* parser,
    json_reason_t reason,
    json_type_t type,
    const json_union_t* un,
    void* cd_)
{
    callback_data_t* cd = (callback_data_t*)cd_;

    /* Print commas */
    if (reason != JSON_REASON_END_ARRAY && reason != JSON_REASON_END_OBJECT &&
        cd->comma)
    {
        cd->comma = 0;
        fprintf(cd->stream, ",");
    }

    /* Decrease depth */
    if (reason == JSON_REASON_END_OBJECT || reason == JSON_REASON_END_ARRAY)
    {
        cd->depth--;
    }

    /* Print newline */
    if (cd->newline)
    {
        cd->newline = 0;
        fprintf(cd->stream, "\n");
        _indent(cd->depth, cd);
    }

    switch (reason)
    {
        case JSON_REASON_NONE:
        {
            /* Unreachable */
            break;
        }
        case JSON_REASON_NAME:
        {
            _print_string(un->string, cd);
            fprintf(cd->stream, ": ");
            cd->comma = 0;
            break;
        }
        case JSON_REASON_BEGIN_OBJECT:
        {
            cd->depth++;
            cd->newline = 1;
            cd->comma = 0;
            fprintf(cd->stream, "{");
            break;
        }
        case JSON_REASON_END_OBJECT:
        {
            cd->newline = 1;
            cd->comma = 1;
            fprintf(cd->stream, "}");
            break;
        }
        case JSON_REASON_BEGIN_ARRAY:
        {
            cd->depth++;
            cd->newline = 1;
            cd->comma = 0;
            fprintf(cd->stream, "[");
            break;
        }
        case JSON_REASON_END_ARRAY:
        {
            cd->newline = 1;
            cd->comma = 1;
            fprintf(cd->stream, "]");
            break;
        }
        case JSON_REASON_VALUE:
        {
#if 0
            json_dump_path(parser);
#endif
            cd->newline = 1;
            cd->comma = 1;
            _print_value(type, un, cd);
            break;
        }
    }

    /* Final newline */
    if (reason == JSON_REASON_END_OBJECT || reason == JSON_REASON_END_ARRAY)
    {
        if (cd->depth == 0)
            fprintf(cd->stream, "\n");
    }

    return JSON_OK;
}

static void _trace(
    json_parser_t* parser,
    const char* file,
    unsigned int line,
    const char* func,
    const char* message)
{
    (void)parser;
    fprintf(stderr, "TRACE: %s(%u): %s(): %s\n", file, line, func, message);
}

static void _parse(const char* path)
{
    json_parser_t parser;
    char* data;
    size_t size;
    json_result_t r;
    callback_data_t cd;
    static json_allocator_t allocator = {
        malloc,
        free,
    };

    cd.depth = 0;
    cd.newline = 0;
    cd.comma = 0;

    if (myst_load_file(path, (void**)&data, &size) != 0)
    {
        fprintf(stderr, "%s: failed to access '%s'\n", arg0, path);
        exit(1);
    }

    if (!(cd.stream = open_memstream(&cd.ptr, &cd.size)))
    {
        fprintf(stderr, "%s: open_memstream() failed\n", arg0);
        exit(1);
    }

    const json_parser_options_t options = {1};

    if ((r = json_parser_init(
             &parser, data, size, _callback, &cd, &allocator, &options)) !=
        JSON_OK)
    {
        fprintf(stderr, "%s: json_parser_init() failed: %d\n", arg0, r);
        exit(1);
    }

    parser.trace = _trace;

    if ((r = json_parser_parse(&parser)) != JSON_OK)
    {
        fprintf(stderr, "%s: json_parser_parse() failed: %d\n", arg0, r);
        exit(1);
    }

    if (cd.depth != 0)
    {
        fprintf(stderr, "%s: unterminated objects\n", arg0);
        exit(1);
    }

    fclose(cd.stream);

    printf("%s", cd.ptr);
}

int main(int argc, char** argv)
{
    arg0 = argv[0];

    if (argc != 2)
    {
        fprintf(stderr, "Usage: %s path\n", argv[0]);
        exit(1);
    }

    _parse(argv[1]);

    return 0;
}
